#!/usr/bin/python3
from dsf.connections import InterceptConnection, InterceptionMode
from dsf.object_model.messages import MessageType
from printer import Printer
from calib import LaserCalib
from controller import HeightController
from export import export_data
import traceback


def success(msg=None):
    intercept.resolve_code(MessageType.Success, msg)


def error(msg=None):
    intercept.resolve_code(MessageType.Error, msg)


def require_param(letter):
    param = code.parameter(letter)
    if param is None:
        raise Exception(f"Expected '{letter}' parameter")
    return param


# Main calibration script
if __name__ == "__main__":
    try:
        # Create an interception connection to catch the M-codes associated with the automatic calibration
        filters = [
            'M1100', 'M1101',                       # Calibration
            'M1110', 'M1111',                       # Height profile
            'M1120', 'M1121', 'M1122', 'M1125',     # Measurement and PID control
            'M1130'                                 # Logging
        ]
        intercept = InterceptConnection(InterceptionMode.PRE, filters=filters)
        intercept.connect()
        control = HeightController()

        while True:
            code = intercept.receive_code()
            intercept.flush(code.channel)
            printer = Printer(intercept)
            
            try:
                if code.majorNumber == 1100:
                    calib = LaserCalib(printer)
                    
                    # Make a PLA cross at the center of the bed
                    create_cross = code.parameter('C')
                    if create_cross is None or create_cross.as_int() != 0:
                        calib.make_cross()

                    # Calibrate the laser
                    calib.calibrate_laser()

                    # Remove the cross
                    remove_cross = code.parameter('R')
                    if remove_cross is None or remove_cross.as_int() != 0:
                        calib.remove_cross()
                        
                    success()

                elif code.majorNumber == 1101:
                    printer._laser_offset = None # Remove the laser offset from the cache, forcing a reload from the file
                    offset = printer.laser_offset
                    
                    X = code.parameter('X')
                    if X is not None:
                        offset[0] += X.as_float()
                    Y = code.parameter('Y')
                    if Y is not None:
                        offset[1] += Y.as_float()
                    Z = code.parameter('Z')
                    if Z is not None:
                        offset[2] += Z.as_float()
                    printer.laser_offset = offset
                    success(f'Laser offset is {offset}')

                elif code.majorNumber == 1110:
                    Z = require_param('Z').as_float()
                    
                    V = code.parameter('V')
                    if V is None:
                        control.volume_profile_reset(Z)
                    else:
                        control.volume_profile_append(Z, V.as_float())
                    success()

                elif code.majorNumber == 1111:
                    Z = require_param('Z').as_float()
                    success(f'Volume profile at height {Z}: {control.volume_profile_sample(Z)}')

                elif code.majorNumber == 1120:
                    if len(code.parameters) == 0:
                        control.next_measurement_reset()
                        success()
                        continue
                    
                    X = require_param('X').as_float_array()
                    Y = require_param('Y').as_float_array()
                    control.next_measurement_append([X[0], Y[0]], [X[1], Y[1]])
                    success()

                elif code.majorNumber == 1121:
                    if len(code.parameters) == 0:
                        control.measurements_reset()
                        success()
                        continue

                    Z = require_param('Z').as_float()
                    control.next_measurement_perform(Z, printer)
                    success()

                elif code.majorNumber == 1122:
                    if len(code.parameters) == 0:
                        success(f'Extrusion factor from PID: {100*control.factor:.0f}%')
                        continue
                    
                    T = require_param('T').as_int_array()
                    S = code.parameter('S')
                    control.set_extrusion_factor(printer, T, None if S is None else S.as_float() / 100)
                    success()

                elif code.majorNumber == 1125:
                    success(f'Laser value: {printer.get_laser_mm()}')

                elif code.majorNumber == 1130:
                    create_graph = code.parameter('C')
                    export_data(control, printer, create_graph is not None and create_graph.as_int() == 1)
                    success()

                else:
                    intercept.ignore_code()
            except Exception as e:
                traceback.print_exc()
                error(str(e))
    finally:
        intercept.close()
